*============================================================================
* mcmccheck:  Residual plots for model checking of an MCMC analysis
* Author:    John Thompson
* Date:      Dec 2012
*============================================================================

program mcmccheck
   version 12.1
   syntax ,   [             ///
   DFilename(string)        /// file containing the data
   Data(string)             /// variable to use in dfilename
   PFilename(string)        /// file containing values from the predictive distribution
   Predictions(string)      /// variable to use in pfilename
   PLot(string)             /// type of plot to produce
   Yaxis(string)            /// what to plot on the y-axis
   Xaxis(string)            /// what to plot on the x-axis
   GOPTions(string asis)    /// options to pass to graphing command
   CGOPTions(string asis)   /// options to pass to graph combine (used for plot=summary)
   Save(string)             /// .dta file for saving the residuals etc.
   KOPTions(string asis)    /// options to pass to kdensity
   BY(string)               /// control symbols on a scatter/lowess plot
   LCondition(string asis)  /// condition on labelling of scatter points
   LVariable(string)        /// variable for labelling of scatter points
   NSamples(integer 0)      /// number of samples in epp
   ]
   
   preserve
   
*--------------------------------
* Check data
*  single value or a variable
*--------------------------------
   if "`dfilename'" != "" &  "`data'" != "" {
      confirm file `dfilename'
      use "`dfilename'", clear
      cap confirm numeric variable `data' , exact
      if _rc != 0 {
         local type "oneway"
         local value = `data'
         if real("`value'") == . {
            di as err "value `data' missing"
            exit(499)
         }
      }
      else local type "twoway"
   }
* --------------------------------------
* Plot type 
* --------------------------------------
   if "`plot'" == "" & "`type'" == "oneway" local plot "histogram"
   if "`plot'" == "" & "`type'" == "twoway" local plot "scatter"
   
   local plots "scatter lowess epp boxplot summary histogram density none"
   local types "twoway twoway twoway twoway toway oneway oneway twoway"
   local found = 0
   local k = 0
   foreach p of local plots {
      local ++k
      if "`plot'" == substr("`p'",1,length("`plot'")) {
         local plot = "`p'"
         local found = 1
         local type : word `k' of `types'
      }
   }
   if `found' == 0 {
      di as err "Plot type `plot' not recognised"
      exit(499)
   }
   if "`plot'" == "none" & "`save'" == "" {
      di as err "no plot produced and residuals not saved: seems pointless"
   }
* --------------------------------------
* SELECT PLOT
* --------------------------------------
   if "`type'" == "oneway" {
      if( "`plot'" == "histogram" ) {
         mcmccheck_histogram , y(`predictions') value(`value') ///
         dfile(`pfilename') gopt(`goptions') kopt(`koptions')
      }
      else if( "`plot'" == "density" ) {
         mcmccheck_histogram , y(`predictions') value(`value') ///
         dfile(`pfilename') gopt(`goptions')  kopt(`koptions') density
      }
   }
* -------------------------
* TWOWAY plots
* -------------------------
   else {
      if "`save'" == "" {
         tempfile mfile
         local mfile "`mfile'.dta"
         local repl "replace"
      }
      else {
         tokenize `"`save'"' , parse(",")
         local mfile `"`1'"'
         local repl "`3'"
      }
      if "`xaxis'" == "" local xaxis "fit"
      if "`yaxis'" == "" local yaxis "res"
* --------------------------------------
* axis OPTION
* --------------------------------------
      local stattypes "residual absresidual ppp tppp rps fit"
      local lab1 "Residual"
      local lab2 "Absolute Residual"
      local lab3 "Posterior predicitve p-value"
      local lab4 "Two-tail posterior predictive p-value"
      local lab5 "Relative Predictive Surprise"
      local lab6 "Mean Prediction"
      local j = 0
      foreach s of local stattypes {
         local ++j
         if "`yaxis'" == substr("`s'",1,length("`yaxis'")) {
            local yaxis  "`s'"
            local ytitle "`lab`j''"
         }
         if "`xaxis'" == substr("`s'",1,length("`xaxis'")) {
            local xaxis  "`s'"
            local xtitle "`lab`j''"
         }
      }
      if "`xtitle'" == "" local xtitle "`xaxis'"
      if "`ytitle'" == "" local ytitle "`yaxis'"
      
* -------------------------
* Measures of Surprise
* -------------------------
      mcmccheck_measures, dfilename(`dfilename') pfilename(`pfilename') ///
      save(`mfile',`repl') obs(`data') pred(`predictions')  kopt(`koptions')
* -------------------------
* Plots
* -------------------------
      if( "`plot'" == "scatter" ) {
         mcmccheck_scatter , mfilename(`mfile')  ofile(`dfilename')  ///
         yaxis(`yaxis') xaxis(`xaxis') by(`by') lvar(`lvariable') ///
         lcond(`lcondition') gopt(ytitle(`ytitle') xtitle(`xtitle') `goptions')
      }
      else if( "`plot'" == "lowess" ) {
         mcmccheck_scatter , mfilename(`mfile')  ofile(`dfilename')  ///
         yaxis(`yaxis') xaxis(`xaxis') by(`by') lvar(`lvariable') ///
         lcond(`lcondition') gopt(ytitle(`ytitle') xtitle(`xtitle') `goptions') lowess
      }
      else if( "`plot'" == "epp" ) {
         if `nsamples' == 0 local nsamples = 50
         mcmccheck_epp , mfilename(`mfile') pfilename(`pfilename') ///
         predict(`predictions') gopt(`goptions') nsamples(`nsamples')
      }
      else if( "`plot'" == "boxplot" ) {
         mcmccheck_boxplot , mfilename(`mfile') ofile(`dfilename') ///
         yaxis("`yaxis'") gopt(`goptions')
      }
      else if( "`plot'" == "summary" ) {
         tempfile f1 f2 f3 f4
         mcmccheck_scatter , mfilename(`mfile') ofile(`dfilename') ///
         yaxis(residual) xaxis(fit) gopt(xtitle(`xtitle') ytitle(Residual) `goptions'  nodraw)
         qui graph save `f1'
         mcmccheck_scatter , mfilename(`mfile')  ofile(`dfilename') ///
         yaxis(absresidual) xaxis(fit) lowess gopt(xtitle(`xtitle') ytitle(Absolute Residual) `goptions' nodraw)
         qui graph save `f2'
         mcmccheck_epp , mfilename(`mfile') pfilename(`pfilename') ///
         predict(`predictions') gopt(`goptions' nodraw) nsamples(50)
         qui graph save `f3'
         use `mfile' , clear
         local b = int(max(10,min(sqrt(_N), 10*ln(_N)/ln(10))))
         mcmccheck_histogram , y(res) ///
         dfile(`mfile') gopt( bin(`b') xtitle(Residual) `goptions' nodraw)
         qui graph save `f4'
         graph combine "`f1'" "`f2'" "`f3'" "`f4'", `cgoptions'
         
      }
   }
   restore
end

*=======================================
* MEASURES - calc measures of surprise
*=======================================
program mcmccheck_measures
   syntax , dfilename(string) ///
   pfilename(string) ///
   save(string) ///
   obs(string) ///
   pred(string) ///
   [ KOPTions(string asis) ]
   
*--------------------------------
* Read real data & save in locals
*--------------------------------
   use "`dfilename'" , clear
   local n = _N
   local k = 0
   forvalues i = 1/`n' {
      local ++k
      local obs`k' = `obs'[`i']
   }
   
*--------------------------------
* save filename for results
*--------------------------------
   tokenize "`save'", parse(",")
   local mfile "`1'"
   local repl  "`3'"
* -------------------------
* Prepare Surprise Measures
* -------------------------
   use "`pfilename'", clear
   tempvar y
   qui gen `y' = 0
   cap postclose pf
   qui postfile pf res absres fit ppp tppp rps using `mfile' , `repl'
   local res = .
   local fit = .
   local ppp = .
   local tppp = .
   local rps = .
   local absres = .
   tempfile tf
   forvalues i = 1/`n' {
      qui su `pred'`i'
      local res = (`obs`i''-r(mean))/r(sd)
      local fit = r(mean)
      local absres = abs(`res')
      qui count if `pred'`i' != .
      local ny = r(N)
      qui count if `pred'`i' > `obs`i''
      local s = r(N)
      qui count if `pred'`i' == `obs`i''
      local ppp = (`s'+0.5*r(N))/`ny'
      local tppp = 2*min(`ppp',1-`ppp')
      tempvar xx yy d
      qui kdensity `pred'`i' , nograph gen(`xx' `yy') `koptions'
      qui su `yy'
      local ymax = r(max)
      qui save "`tf'.dta", replace
      qui gen `d' = abs(`xx'-`obs`i'')
      sort `d'
      local h = `yy'[1]
      use "`tf'.dta" , clear
      local rps = 100*`h'/`ymax'
      drop `xx' `yy'
      post pf (`res') (`absres') (`fit') (`ppp') (`tppp') (`rps')
   }
   
   postclose pf
   use `mfile' , clear
   
   label var res "Residual"
   label var absres "Absolute Residual"
   label var fit "Mean Prediction"
   label var ppp "Posterior predictive p-value"
   label var tppp "Two-tail posterior predictive p-value"
   label var rps "Relative predicitve surprise"
   qui save `mfile' , replace
end

*=======================================
* SCATTER PLOT
*=======================================
program mcmccheck_scatter
   syntax , [               ///
   mfilename(string)             ///  file in which measures are stored
   OFile(string)            ///
   YAxis(string)            ///
   XAxis(string)            ///
   LVariable(string)        ///
   LCondition(string asis)  ///
   LOWESS                   ///
   BY(string)               ///
   GOPTions(string asis) ]
   
* -------------------------
* merge data & measures
* -------------------------
   qui use "`mfilename'" , clear
   qui merge 1:1 _n using "`ofile'"
   qui drop _merge
* -------------------------
* Construct axes
* -------------------------
   mcmccheck_convert , t("`xaxis'")
   local xax = r(t)
   mcmccheck_convert , t("`yaxis'")
   local yax = r(t)
   local ytitle = r(t)
   if "`yax'" == "res" {
      qui su res, meanonly
      if r(min) < 0 & r(max) > 0 local addline "yline(0)"
      local ytitle "Residual"
   }
   else if "`yax'" == "absres" {
      local ytitle "Absolute Residual"
   }
   else if "`yax'" == "ppp" {
      local addline "yline(0.025 0.975,lpat(dash))"
      local range "yscale(range(0 1)) ylabel(0(.2)1)"
      local ytitle "Posterior predictive p-value"
   }
   else if "`yax'" == "tppp" {
      local addline "yline(0.05,lpat(dash))"
      local range "yscale(range(0 1)) ylabel(0(.2)1)"
      local ytitle "Two-tail posterior predictive p-value"
   }
   else if "`yax'" == "rps" {
      local addline "yline(5,lpat(dash))"
      local range "yscale(range(0 100)) ylabel(0(20)100)"
      local ytitle "Relative Predictive Surprise"
   }
* -------------------------
* Calculate plotting variables
* -------------------------
   tempvar x y
   qui gen `x' = `xax'
   qui gen `y' = `yax'
   sort `x'
   
* -------------------------
* LOWESS plot
* -------------------------
   if "`lowess'" != "" {
      tempvar lfit
      qui lowess `y' `x' , gen(`lfit') nograph
      local lowplot "(line `lfit' `x', clpat(solid))"
   }
* -------------------------
* SCATTER plot
* -------------------------
* -------------------------
* Symbols for levels of by
* -------------------------
   if "`by'" != "" {
      cap confirm numeric variable `by' , exact
      if !_rc {
         qui levelsof `by' , local(lev)
         local scat ""
         foreach lv of local lev {
            local scat "`scat' (scatter `y' `x' if `by' == `lv')"
         }
      }
      else {
         cap confirm string variable `by' , exact
         if !_rc {
            qui levelsof `by' , local(lev)
            local scat ""
            foreach lv of local lev {
               local scat `"`scat' (scatter `y' `x' if `by' == "`lv'")"'
            }
         }
         else {
            di as err "`by' must be a string or numeric variable stored in the data file"
            exit(499)
         }
      }
   }
* -------------------------
* One plotting symbol
* -------------------------
   else {
      local scat "(scatter `y' `x')"
   }
* -------------------------
* Set labels
* -------------------------
   if `"`lcondition'"' != "" | "`lvariable'" != "" {
      tempvar id
      if "`lvariable'" == "" qui gen `id' = _n
      else qui gen `id' = `lvariable'
      mcmccheck_convert , t("`lvariable'")
      local lvariable = r(t)
      mcmccheck_convert , t(`"`lcondition'"')
      local lcondition = r(t)
      local labs "(scatter `y' `x' if `lcondition' , msym(none) mlab(`id'))"
   }
* -------------------------
* Make plot
* -------------------------
   twoway `scat' `lowplot' `labs', `addline' `goptions' ///
   ytitle("`ytitle'") legend(off)  `range'
end

*--------------------------------
* Convert formulae
*--------------------------------
program mcmccheck_convert, rclass
   syntax , [ t(string asis) ]
   
   tokenize `"`t'"', parse("()+-*/^,<>[] ")
   local t ""
   local i = 1
   while `"``i''"' != "" {
      if `"``i''"' == substr("residuals",1,max(3,strlen(`"``i''"'))) {
         local t `"`t'res"'
      }
      else if `"``i''"' == substr("absresiduals",1,max(4,strlen(`"``i''"'))) {
         local t `"`t'absres"'
      }
      else {
         local t `"`t'``i''"'
      }
      local i = `i' + 1
   }
   return local t `"`t'"'
end

*=======================================
* HISTOGRAM
*=======================================
program mcmccheck_histogram
   syntax , Y(string) ///
   dfile(string) ///
   [ value(string) ///
   GOPTions(string asis) ///
   KOPTions(string asis) ///
   density ]
   
   use "`dfile'", clear
   tempvar xx yy d
   
*-------------------------------
* Surprise at the observation
*-------------------------------
   if "`value'" != "" {
      qui count if `y' != .
      local n = r(N)
      qui count if `y' > `value'
      local ppp = r(N)
      qui count if `y' == `value'
      local ppp = (`ppp'+0.5*r(N))/`n'
      local ppp : di %8.4f `ppp'
      qui su `y'
      local rd : di %7.3f (`value'-r(mean))/r(sd)
      qui kdensity `y' , nograph gen(`xx' `yy') `koptions'
      tempfile tf
      qui save "`tf'.dta" , replace
      qui gen `d' = abs(`xx'-`value')
      sort `d'
      local h = `yy'[1]
      use "`tf'.dta", clear
      qui su `yy'
      local ymax = r(max)
      local rs : di %6.2f 100*`h'/`ymax' "%"
      local bin = max(10,int(sqrt(`n')))
      local xline "xline(`value')"
      local note `"note("Predictive p = `ppp'" "RPS = `rs'" "residual = `rd'")"'
   }
*-------------------------------
* make the plot
*-------------------------------
   if "`density'" == "" {
      qui histogram `y' , percent `xline' ///
      ytitle(Percent) `note' xtitle(`y') `goptions'
   }
   else {
      if "`value'" == "" qui kdensity `y' , nograph gen(`xx' `yy') `koptions'
      qui twoway line `yy' `xx' , `xline' ///
      ytitle("") `note' xtitle(`y') `goptions'
   }
end

*=======================================
* EPP - empirical probability plot
*=======================================
program mcmccheck_epp
   syntax , ///
   predict(string)       ///
   mfilename(string)          ///
   pfilename(string)     ///
   nsamples(integer)     ///
   [ GOPTions(string asis) ]
   
   
   tempfile epp pd tf sepp
*-----------------------------
* Order & Save real residuals
*-----------------------------
   use "`mfilename'" , clear
   local n = _N
   keep res
   sort res
   qui save "`epp'" , replace
   
*-----------------------------
* standardise predictions
*-----------------------------
   use "`pfilename'", clear
   forvalue i = 1/`n' {
      tempvar res`i'
      qui su `predict'`i'
      qui gen `res`i'' = (`predict'`i'-r(mean))/r(sd)
   }
   qui save `pd' , replace
   
*-----------------------------
* Take repeated random rows from
* standardised predictions
*-----------------------------
   local start = 1 + int((_N/`nsamples')*runiform())
   forvalues k = 1/`nsamples' {
      local row = `start'+int((`k'-1)*_N/20)
      use "`pd'" , clear
      qui postfile pf r`k' using "`tf'", replace
      forvalue i = 1/`n' {
         post pf (`res`i''[`row'])
      }
      postclose pf
      use "`tf'", clear
      sort r`k'
      if `k' > 1 {
         qui merge 1:1 _n using "`sepp'"
         qui drop _merge
      }
      qui save "`sepp'", replace
   }
   
*-----------------------------
* Make the plot
*-----------------------------
   use "`sepp'", clear
   xpose , clear
   local k = 0
   foreach v of varlist * {
      local ++k
      qui centile `v' ,cent(2.5 97.5)
      local lb`k' = r(c_1)
      local ub`k' = r(c_2)
      qui su `v', meanonly
      local m`k' = r(mean)
   }
   use `epp' , clear
   qui gen m = .
   qui gen ub = .
   qui gen lb = .
   local n = _N
   forvalues k=1/`n' {
      qui replace m = `m`k'' in `k'
      qui replace lb = `lb`k'' in `k'
      qui replace ub = `ub`k'' in `k'
   }
   twoway (scatter res m) (line m m, clpat(solid) lcol(blue)) ///
   (line ub m , clpat(dash) lcol(blue)) (line lb m, clpat(dash) lcol(blue)) , ///
   xtitle("Expected `mLabel'") ytitle("Observed `mLabel'") legend(off) `goptions'
end


*=======================================
* BOXPLOT
*=======================================
program mcmccheck_boxplot
   syntax , mfilename(string) ///
   yaxis(string) ///
   OFILE(string)  ///
   [ GOPTions(string asis) ]
   
   
   qui use "`mfilename'" , clear
   qui merge 1:1 _n using "`ofile'"
   qui drop _merge
   mcmccheck_convert , t("`yaxis'")
   local xt = r(t)
   tempvar y
   qui gen `y' = `xt'
   graph box `y' , ytitle(`yaxis') `goptions'
end

