*============================================================================
* mhstrnc: Metropolis-Hastings Sampler based on a normal proposal distribution applied
*          to a finite range (a,b) using the transformed scale log[(y-a)/(b-y)]
* Author:  John Thompson
* Date:    Dec 2012
*============================================================================

program mhstrnc, rclass
   version 12.1
   syntax anything  [fweight] [if] [in], ///
      SD(string)          ///
	  [                   ///
	  Lbound(real -1e20)  ///
	  Ubound(real  1e20)  ///
	  LOGP(real 1e20)     ///
	  DEBUG               ///
	  ]

   tokenize "`anything'"
   local logpost "`1'"
   local b "`2'"
   local ipar = `3'

   tempname logl newlogl scale1 scale2 newvalue oldvalue logalpha u
   if "`debug'" == "" local qui "qui"
   scalar `logl' = .
   scalar `newlogl' = .
   scalar `oldvalue' = `b'[1,`ipar']

   if `lbound' < -1e19 local lbound "."
   else if `b'[1,`ipar'] < `lbound' {
      di as err "Parameter `ipar': initial value outside the bounds"
	  exit(499)
   }
   if `ubound' > 1e19 local ubound "."
   else if `b'[1,`ipar'] > `ubound' {
      di as err "Parameter `ipar': initial value outside the bounds"
	  exit(499)
   }

   if `logp' < 1e19 scalar `logl' = `logp'
   if `logl' == .  {
      `qui' `logpost' `logl' `b' `ipar' [`weight'`exp'] `if' `in'
      if `logl' == .  {
         di as err "mhstrnc: cannot calculate log-posterior for parameter `ipar'"
         matrix list `b'
         exit(1400)
      }
   }
   if `lbound' == . {
      if `ubound' == . {
* [.,.] untransformed scale
         local delta = `sd'*rnormal()
         scalar `newvalue' = `oldvalue' + `delta'
         scalar `scale1' = 0
         scalar `scale2' = 0
      }
      else {
* [.,u] reverse log-transformed scale log(u-y)
   if `oldvalue' >= `ubound' {
      di as err "Parameter `ipar': initial value outside bounds"
	  exit(499)
   }
         local delta = exp( `sd'*rnormal())
         scalar `newvalue' = `ubound' - (`ubound' - `oldvalue')/`delta'
         scalar `scale1' = log(`ubound'-`oldvalue')
         scalar `scale2' = log(`ubound'-`newvalue')
      }
   }
   else {
      if `ubound' == . {
* [l,.] log-transformed scale log(y-l)
   if `oldvalue' <= `lbound' {
      di as err "Parameter `ipar': initial value outside bounds"
	  exit(499)
   }
         local delta = exp( `sd'*rnormal())
         scalar `newvalue' = (`oldvalue'-`lbound')*`delta'+`lbound'
         scalar `scale1' = log(`oldvalue'-`lbound')
         scalar `scale2' = log(`newvalue'-`lbound')
      }
      else {
* [l,u] logistically-transformed scale log[(y-l)/(u-y)]
   if `oldvalue' <= `lbound' | `oldvalue' >= `ubound' {
      di as err "Parameter `ipar': initial value outside bounds"
	  exit(499)
   }
         local delta = exp( `sd'*rnormal())
         scalar `newvalue' = (`ubound'*(`oldvalue'-`lbound')*`delta'+`lbound'*(`ubound'-`oldvalue'))/(`ubound'-`oldvalue'+(`oldvalue'-`lbound')*`delta')
         scalar `scale1' = log(`ubound'-`oldvalue')+log(`oldvalue'-`lbound')
         scalar `scale2' = log(`ubound'-`newvalue')+log(`newvalue'-`lbound')
      }
   }
   matrix `b'[1,`ipar'] = `newvalue'
   `qui' `logpost' `newlogl' `b' `ipar' [`weight'`exp'] `if' `in'
   if `newlogl' == .  {
      di as err "mhstrnc: cannot calculate log-posterior for parameter `ipar'"
      matrix list `b'
      exit(1400)
   }
   
   
   scalar `logalpha' = (`newlogl'+`scale2') - (`logl'+`scale1')
   if log(runiform()) < `logalpha'   local accept = 1
   else {
      local accept = 0
      matrix `b'[1,`ipar'] = `oldvalue'
      scalar `newlogl' = `logl'
   }
   return scalar accept = `accept'
   return scalar logp = `newlogl'
end
